<?php

namespace Drupal\mcapi_credcom\Controller;

use Drupal\mcapi\Entity\Wallet;
use Drupal\mcapi\Entity\Transaction;
use Drupal\mcapi_credcom\FinalExceptionSubscriber;
use Drupal\mcapi_credcom\CreditCommons\Workflow;
use Drupal\mcapi_credcom\Functions;
use CreditCommons\Exceptions\CCFailure;
use CreditCommons\Exceptions\WorkflowViolation;
use CreditCommons\TransactionInterface as CCTransactionInterface;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\TempStore\SharedTempStore;

/**
 * Handle all incoming requests from the trunkward node.
 */
class EndPoint extends \Drupal\Core\Controller\ControllerBase {

  private $logger;
  private $config;
  private $tempstore;

  /**
   * See HashChecker for more set up operations.
   */
  public function __construct(Request $request, SharedTempStore $credcom_tempstore) {
    global $error_context;
    $this->config = $this->config('mcapi_credcom.settings');
    $this->logger = $this->getLogger('mcapi_credcom');
    $this->tempstore = $credcom_tempstore;
    Functions::setErrorContext($this->config);
    // Generate custom error responses for exceptions inside Drupal.
    \Drupal::getContainer()->set('exception.final', new FinalExceptionSubscriber(\Drupal::configFactory()));
    $this->logger->info('@request', ['@request' => '<pre>'.print_r($request, 1).'</pre>']);
  }

  /**
   * @param ContainerInterface $container
   * @return static
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('request_stack')->getCurrentRequest(),
      $container->get('tempstore.shared')->get('mcapi_credcom')
    );
  }

  /**
   * Having validated with the hash, return that this leaf node is online.
   */
  function handshake() {
    // Empty array means there are no further leafward nodes.
    return JsonResponse::fromJsonString('{"data": []}');
  }

  /**
   * Return the exchange rate.
   */
  function convert() {
    $result = [
      'data' => [
        'symbol' => $this->config->get('local_curr_symbol'),
        'rate' => 1 * $this->config->get('trunkward.rate')
      ]
    ];
    return JsonResponse::fromJsonString(json_encode($result));
  }

  /**
   * @param Request $request
   * @return JsonResponse
   */
  function autocompleteNames(Request $request) {
    $result = [];
    $fragment = $request->query->get('acc_path');
    if ($fragment and $mode = $this->config->get('privacy.acc_path_autocomplete')) {
      $q = $this->entityTypeManager()->getStorage('mc_wallet')->getQuery()
        ->accessCheck(TRUE)
        ->range(0, 10);
      if(is_numeric($fragment)) {
        $q->condition('wid', $fragment.'%', 'LIKE');
      }
      elseif ($mode == CREDCOM_AUTOCOMPLETE_NAME) {
        $q->condition('name', '%'.$fragment.'%', 'LIKE');
      }
      else {
        $q->condition('wid', 0); // No results.
      }
      $wids = $q->execute();
      if ($mode == CREDCOM_AUTOCOMPLETE_NAME) {
        foreach (Wallet::loadMultiple($wids) as $wid => $wallet) {
          // Credit commons wallet names can't have spaces.
          $result[] = str_replace(' ', '-', strtolower($wallet->name->value));
        }
      }
      else {
        $result = $wids;
      }
    }
    return JsonResponse::fromJsonString(json_encode(['data' => $result]));
  }

  /**
   * Handle incoming transaction request.
   * @return JsonResponse
   */
  function relayTransaction(Request $request) {
    $contents = $request->getContent();
    $this->logger->debug('Received json @credcom', ['@credcom' => $contents]);
    $json = json_decode($contents);
    $json_entries = (array)$json->entries;
    $count_entries = count($json_entries);

    $transaction_props = [
      'uuid' => $json->uuid,
      'workflow' => $json->type,
      'entries' => [Functions::convertJsonToMcapiEntry(array_shift($json_entries))],
    ];

    $transaction = Transaction::create($transaction_props);
    foreach ($json_entries as $e) {
      $transaction->addEntry(Functions::convertJsonToMcapiEntry($e));
    }
    foreach ($transaction->validate() as $violation) {
      $message = get_class($violation) .' error in new transaction: '. $violation->getMessage();
      \Drupal::logger('mcapi_credcom')->debug('@message', ['@message' => $message]);
      // @todo work out the right kind of CCError.
      throw new CCFailure($message);
    }
    if ($transaction->workflow->entity->creation['confirm']) {
      $transaction->set('state', CCTransactionInterface::STATE_VALIDATED);
      $this->tempstore->set($transaction->uuid(), $transaction);
      $response_code = 200;
    }
    else {
      $transaction->save();
      $response_code = 201;
    }
    $new_entries = array_map(
      function ($e) {return Functions::jsonSerializeEntry($e);},
      array_slice(iterator_to_array($transaction->entries), $count_entries)
    );
    // These entries need converting.
    return JsonResponse::fromJsonString(json_encode(['data' => $new_entries], $response_code));
  }

  /**
   *
   * @param string $uuid
   * @param string $target_state
   * @return JsonResponse
   * @throws WorkflowViolation
   */
  function changeState(string $uuid, string $target_state) {
    $transaction = $this->tempstore->get($uuid); // already assembled and validated
    if (!$transaction) {
      $txs = $this->entityTypeManager()->getStorage('mcapi_transaction')->loadByProperties(['uuid' => $uuid]);
      if (empty($txs)) {
        throw new CCFailure("Transaction $uuid not found in Drupal tempStore.");
      }
      $transaction = reset($txs);
    }
    $workflow = new Workflow($transaction);
    $wallet_id = \Drupal::config('mcapi_credcom.settings')->get('wallet_id');

    if (!$workflow->canTransitionToState($wallet_id, $target_state)) {
      throw new WorkflowViolation(type: $transaction->workflow->target_id, from: $transaction->state->value, to: $target_state);
    }

    $transaction->set('state', $target_state);

    $transaction->save();
    $this->tempstore->delete($uuid);
    return JsonResponse::fromJsonString('', 201);
  }


}
