<?php

namespace Drupal\mcapi_credcom;

use CreditCommons\Exceptions\CCFailure;
use CreditCommons\Exceptions\CCError;
use CreditCommons\Exceptions\DoesNotExistViolation;
use Drupal\mcapi\Entity\Transaction;
use Drupal\mcapi\Entity\Storage\WalletStorage;
use Drupal\Core\Render\Markup;
use Drupal\Core\Config\Config;
use Drupal\mcapi_forms\Entity\McForm;

/**
 * Access check for requests from trunkwards nodes.
 */
class Functions {

  static $remoteAccPath;

  /**
   * @param Transaction $transaction
   * @return bool
   *   TRUE if the balance of trade wallet is used in the transaction.
   */
  static function isRemote(Transaction $transaction) : bool {
    $credcom_wid = \Drupal::config('mcapi_credcom.settings')->get('wallet_id');
    return isset($transaction->allWallets()[$credcom_wid]);
  }

  /**
   * Get the wallet ID from the credcom path.
   * @param string $cc_path
   *   path of an account provided by the trunkward node.
   * @return int
   */
  static function convertCCaddressToWalletId(string $cc_path, \stdClass &$metadata) : int {
    $local_acc_name = static::credcomPathIsLocal($cc_path);
    if (!$local_acc_name) {
      $wid = (int)\Drupal::config('mcapi_credcom.settings')->get('wallet_id');
      $metadata->{$wid} = $cc_path;
    }
    elseif (!isset($wid) and is_numeric($local_acc_name)) {
      // Assuming its not a float!
      $wid = intval($local_acc_name);
    }
    if (!isset($wid)){
      $entityTypeManager = \Drupal::entityTypeManager();
      // Wallet name
      $wids = $entityTypeManager->getStorage('mc_wallet')
        ->getQuery()->condition('name', $local_acc_name)
        ->accessCheck(FALSE)
        ->execute();
      if ($wids) {
        $wid = reset($wids);
      }
    }
    if (!isset($wid)) {
      // User email address
      if (filter_var($local_acc_name, FILTER_VALIDATE_EMAIL)) {
        $users = $entityTypeManager->getStorage('user')->loadByProperties(['mail' => $local_acc_name]);
        $user = reset($users);
      }
      // User name
      else {
        $uids = $entityTypeManager->getStorage('user')
          ->getQuery()->condition('name', $local_acc_name)
          ->accessCheck(FALSE)
          ->execute();
        if ($uids) {
          $user = \Drupal\user\Entity\User::load(reset($uids));
        }
      }
      if (isset($user)) {
        $wallet = WalletStorage::firstWalletOfEntity($user);
        if ($wallet) {
          $wid = $wallet->id();
        }
      }
    }
    if (!isset($wid)) {
      throw new DoesNotExistViolation(id: $cc_path, type: 'account');
    }
    return $wid;
  }

  /**
   * Get the last hash which should match the trunkward node's.
   */
  static function getLastHash() {
    return \Drupal::database()
      ->select('mcapi_credcom_hashchain', 'chain')
      ->fields('chain', ['hash'])
      ->orderBy('id', 'desc')
      ->range(0, 1)
      ->execute()
      ->fetchField();
  }

  /**
   * Utility, check if a credcom address refers to a local account, and if so
   * return the localised account name.
   *
   * @param string $cc_path
   * @return string
   *
   */
  static function credcomPathIsLocal(string $cc_path) : string {
    $conf = \Drupal::config('mcapi_credcom.settings');
    $local_name = '';
    $parts = explode('/', $cc_path);
    $pos = array_search($conf->get('cc_user'), $parts);
    if ($pos !== FALSE) {
      $parts = array_slice($parts, $pos+1);
    }
    if (count($parts) == 1) {// Must be the intertrading wallet
      $first_part = reset($parts);
      if ($conf->get('trunkward_node_id') <> $first_part) {
        $local_name = $first_part;
      }
    }
    return $local_name;
  }

  /**
   * Exception handler to return credit commons errors to the network
   * Consider replacing with a class that extends Drupal\Core\Exception\ExceptionHandler
   * @param \Throwable $ex
   */
  static function CredcomExceptionHandler(\Throwable $ex) {
    $err = CCError::convertException($ex);
    \Drupal::logger('mcapi_credcom')
      ->error('@message', ['@message' => Markup::create('<pre>'.print_r($err, 1).'</pre>')]);
    header('Content-Type: application/json');
    header('HTTP/1.1: 500 Server error');
    print json_encode(['errors' => [$err]]);
    exit;
  }

  /**
   * Error handler to return credit commons errors to the network
   * @param \Throwable $ex
   */
  static function CredcomErrorHandler($severity, $message, $filename, $lineno) {
    $trace = array_slice(debug_backtrace(DEBUG_BACKTRACE_IGNORE_ARGS), 1);
    if ($severity > 8000) {// deprecation messages
      \Drupal::logger('mcapi_credcom')
      ->warning(
        'Deprecated: @message\n@break\n@trace',
        ['@message' => $message, '@break'=> "$filename:$lineno", '@trace' => Markup::Create('<pre>'.print_r($trace, 1).'</pre>')]
      );
    }
    else {
      $err = new CCFailure($message);
      $err->trace = $trace;
      $err->break = $filename . ':'. $lineno;
      \Drupal::logger('mcapi_credcom')
        ->error('@message', ['@message' =>  Markup::create('<pre>'.print_r($err, 1).'</pre>')]);
      header('Content-Type: application/json');
      header('HTTP/1.1: 500 Server error');
      print json_encode(['errors' => [$err]], 500);
      exit;
    }
  }

  /**
   * Edit the ledger and change the transaction workflows retrospectively
   * @param string $old_id
   * @param string $new_id
   */
  public static function updateTransactionTypes(string $old_id, string $new_id) {
    \Drupal::database()->update('mc_transaction')
      ->fields(['workflow' => $new_id])
      ->condition('workflow', $old_id)
      ->execute();
    \Drupal::database()->update('mc_transactions_index')
      ->fields(['workflow' => $new_id])
      ->condition('workflow', $old_id)
      ->execute();
    \Drupal::logger('credcom')->notice(
      'Edited the ledger changing transaction workflow @old for @new',
      ['@old' => $old_id, '@new' => $new_id]
    );
    if (class_exists('Drupal\mcapi_forms\Entity\McForm')) {
      foreach (McForm::loadMultiple() as $id => $mc_form) {
        if ($mc_form->getworkflowId() == $old_id) {
          $mc_form->setWorkflow($new_id)->save();
        }
      }
    }
  }

  /**
   * Change the name of a local transaction workflow
   * @param Workflow $mcType
   *
   * @todo this could be moved to TrunkwardNodeRequester?
   */
  public static function localiseTransactionTypeName(Workflow $mcType) {
    $newWorkflow = clone($mcType);
    $newWorkflow->uuid = new \Drupal\Component\Uuid\Uuid();
    $newWorkflow->id = $mcType->id() . '_locl';
    $newWorkflow->local = TRUE;
    $newWorkflow->save();
    $mcType->delete();
    \Drupal::logger('credcom')->notice(
      'replaced transaction type @old with @new',
      ['@old' => $mcType->id(), '@new' => $newWorkflow->id()]
    );
    static::updateTransactionTypes($mcType->id(), $newWorkflow->id());
  }

  /**
   *
   * @param Transaction $transaction
   */
  static function saveHash(Transaction $transaction) : void {
    if ($transaction->version) {
      // only hash the entries that involve the interrading wallet;
      $wid = \Drupal::config('mcapi_credcom.settings')->get('wallet_id');
      $entries = array_filter(
        iterator_to_array($transaction->entries),
        function($e) use ($wid) { return $e->payee_wid == $wid xor $e->payer_wid == $wid; }
      );
      $hash_string = static::getHashString(
        $transaction->uuid(),
        $transaction->state->value,
        $transaction->version,
        $entries
      );
      \Drupal::database()->insert('mcapi_credcom_hashchain')
        ->fields([
          'uuid' => $transaction->uuid(),
          'version' => $transaction->version,// calculated property
          'hash' => md5($hash_string),
          'hash_string' => $hash_string
        ])
        ->execute();
    }
  }

  /**
   *
   * @param string $uuid
   * @param string $state
   * @param int $version
   * @param array $entries
   *
   * @return string
   *
   * @todo replace this with the cc-php-lib?
   */
  private static function getHashString(string $uuid, string $state, int $version, array $entries) : string {
    $rows = [];
    foreach ($entries as $e) { /** @var EntryItem $e */
      if (empty($e->quant_t)) {
        throw new \Exception("no trunkward quant in entry.");
      }
      $rows[] = $e->quant_t.'|'.$e->description;
    }
    return join('|', [
      Functions::getlastHash(),
      $uuid,
      $state,
      join('|', $rows),
      $version,
    ]);
  }

  /**
   * \CreditCommons needs this global to generate error messages.
   *
   * @global Config $config
   */
  static function setErrorContext(Config $config) : void {
    global $error_context;
    $error_context = (object)[
      'node' => $config->get('cc_user') ?? '',
      'method' => \Drupal::request()->getMethod(),
      'path' => \Drupal::request()->getPathInfo(),
      'user' => $config->get('trunkward_node_id') ?? ''
    ];
  }


  /**
   * return an entry which will become a \Creditcommons\Entry with json_encode
   * @param EntryItem $entry
   * @return array
   */
  public static function jsonSerializeEntry(EntryItem $entry) :array {
    $config = \Drupal::config('mcapi_credcom.settings');
    $payee = $entry->metadata[$this->payee_wid] ?? $config->get('node_name') . '/'.$entry->payee_wid;
    $payer = $entry->metadata[$this->payer_wid] ?? $config->get('node_name') . '/'.$entry->payer_wid;
    unset($entry->metadata[$this->payee_wid],  $entry->metadata[$this->payer_wid]);
    return [
      'payee' => $payee,
      'payer' => $payer,
      'description' => $entry->description,
      'quant' => $entry->quant * $config->get('trunkward.rate'),
      'metadata' => $entry->metadata,
      'primary' => $entry->primary
    ];
  }

  public static function convertJsonToMcapiEntry(\stdClass $entry) : array {
    return [
      'payee_wid' => Functions::convertCCaddressToWalletId($entry->payee, $entry->metadata),
      'payer_wid' => Functions::convertCCaddressToWalletId($entry->payer, $entry->metadata),
      'description' => $entry->description,
      'quant' => $entry->quant,
      'metadata' => (array)$entry->metadata
    ];
  }

}
