<?php

namespace Drupal\mcapi_credcom;

use CreditCommons\BaseTransaction;
use CreditCommons\Workflow as CcWorkflow;
use Drupal\mcapi\Entity\Workflow as McWorkflow;
use Drupal\mcapi_credcom\Functions;
use Drupal\mcapi_forms\Entity\McForm;
use Drupal\Core\Config\ConfigFactory;
use Drupal\Core\Logger\LoggerChannel;
use Drupal\Core\Render\Markup;
use GuzzleHttp\RequestOptions;


/**
 * Service handling outgoing requests to the trunkward node.
 */
class TrunkwardNodeRequester extends \CreditCommons\NodeRequester {

  private $logger;
  private $config;

  /**
   * Constructor
   *
   * @param ConfigFactory $config_factory
   * @param LoggerChannel $logger_channel
   */
  public function __construct(ConfigFactory $config_factory, LoggerChannel $logger_channel) {
    $this->config = $config_factory->get('mcapi_credcom.settings');
    Functions::setErrorContext($this->config);
    $this->logger = $logger_channel;
    if ($this->configured()) {
      // Initiate the service from the settings.
      parent::__construct(
        $this->config->get('trunkward.url'),
        $this->config->get('node_name'),
        Functions::getlastHash()
      );
    }
    $this->catch = TRUE;
  }

  public function configured() : bool {
    return $this->config->get('node_name') and $this->config->get('trunkward.url') and $this->config->get('trunkward.id');
  }

  /**
   * Ensure the local transaction->type reflect the remote workflows.
   *
   * @note workflows should change very rarely and even then should be
   * deprecated long before being removed to allow transactions to complete.
   * There is no procedure for converting transactions with old workflows to new.
   *
   * On update separate wfs into local only, remoteonly and both
   *   foreach localonly
   *     ensure the ->local is true.
   *   foreach remoteonly
   *     if a remote name is the same as a local name
   *       if the config->local is true
   *         rename the local config with local_ prefix && rewrite the ledger replacing the type name.
   *     copy the remote to config with ->local=false
   *   foreach both
   *     ensure the transactionType has local->false
   *     if the names are different
   *       rewrite the ledger to use the remote name.
   * @note this still allows that two workflows might have the same label
   */
  public function updateWorkflows() {
    $clear_cache = FALSE;
    [$remote_wfs, $local_wfs, $common_wfs] = $this->compareWorkflows();
    foreach ($local_wfs as $local_wf) {
      /** @var \Drupal\mcapi\Entity\Type $local_wf */
      // All the items without trunkward equivalents should have the local flag set to TRUE
      $mcType = McWorkflow::load($local_wf->id);
      if (!$mcType->local) {
        $mcType->local = TRUE;
        $mcType->save();
        $clear_cache = TRUE;
        \Drupal::logger('credcom')->notice(
          'Workflow @name is no longer available remotely and is now considered a local workflow',
          ['@name' => $mcType->id]
        );
      }
      $labels[] = $local_wf->label;
    }
    foreach ($remote_wfs as $remote_wf) {
      /** @var CreditCommons\Workflow $remote_wf */
      // All trunkward workflows must be represented by local config.
      if ($mcType = McWorkflow::load($remote_wf->id)) {
        if (!$mcType->local) {
          \Drupal::logger('credcom')->error("Shouldn't be possible to have a remote workflow with and a different hash but the same name as a non-local transaction type.");
        }
        Functions::localiseTransactionTypeName($mcType); // Also updates existing transactions
        \Drupal::logger('credcom')->notice(
          'Workflow @name was the same as a remote workflow and has been renamed to @new_name',
          ['@name' => $remote_wf->id, '@new_name' => $mcType->id()]
        );
      }
      mcapi_credcom_convert_credcom_wf_to_mc_workflow($remote_wf)->save();
      $clear_cache = TRUE;
      \Drupal::logger('credcom')->notice(
        'The remote workflow @name has been saved locally',
        ['@name' => $remote_wf->id]
      );
        $clear_cache = TRUE;
      $labels[] = $remote_wf->label;
    }
    foreach ($common_wfs as $hash => $common_wf) {
      /** @var CreditCommons\Workflow $common_wf */
      // Common wfs must have the same name with the config->local set to false
      $mcType = McWorkflow::load($common_wf->localId);
      if ($common_wf->id <> $mcType->id) {
        Functions::updateTransactionTypes($mcType->id(), $common_wf->id);
        $map[$mcType->id()] = $common_wf->id;
        $mcType->delete();
        $clear_cache = TRUE;
        mcapi_credcom_convert_credcom_wf_to_mc_workflow($common_wf)->save();
        \Drupal::logger('credcom')->notice(
          'Replaced transactionType @type with remote workflow @id which had the same hash.',
          ['@type' => $mcType->id(), '@id' => $common_wf->id]
        );
      }
      elseif ($mcType->local) {
        $mcType->local = FALSE;
        $mcType->save();
        $clear_cache = TRUE;
        \Drupal::logger('credcom')->notice('Transaction type @type was converted to a local version of a remote workflow.');
      }
      $labels[] = $common_wf->label;
    }
    if (count($labels) > count(array_unique($labels))) {
      \Drupal::logger('credcom')->warning('Two workflows share the same label.');
    }
    if ($clear_cache) {
      \Drupal::service('cache_tags.invalidator')->invalidateTags(['entity_type:mc_workflow']);
    }
    // untested!
    if (class_exists('\Drupal\mcapi_forms\Entity\McForm')) {
      foreach (McForm::loadMultiple() as $form) {
        if (isset($map[$form->getWorkflowId()])) {
          $form->setWorkflowId($map[$form->getWorkflowId()])->save();
        }
      }
    }
  }

  /**
   *
   * @return CcWorkflow[][]
   *   Arrays of the remote-only, local-only and common workflows.
   */
  private function compareworkflows() {
    $local_wfs = $localised_wfs = $remote_wfs = [];
    // List the remote and local workflows keyed by hash
    foreach (parent::getWorkflows() as $hash => $wf) {
      $trunkward_wfs[$hash] = $wf;
    }
    foreach (McWorkflow::loadMultiple() as $mcType) {
      $type = new \stdClass;
      foreach (get_object_vars($mcType) as $varname => $val) {
        $type->{$varname} = $val;
      }
      $wf = new CcWorkflow($type);
      $hash = $wf->getHash();
      if (isset($trunkward_wfs[$hash])) {
        $trunkward_wfs[$hash]->localId = $mcType->id();
      }
      $config_wfs[$hash] = $wf;
    }

    return [
      array_diff_key($trunkward_wfs, $config_wfs), // Remote only
      array_diff_key($config_wfs, $trunkward_wfs), // Local only
      array_intersect_key($trunkward_wfs, $config_wfs) // Common
    ];
  }

  /**
   * {@inheritDoc}
   *
   * Theoretically this could be disintermediated by javascript calling the
   * trunkward node directly, but then the creds would have to be sent to client.
   */
  public function accountNameFilter(string $acc_path = '', $limit = 10) : array {
    $result = [];
    try {
      $result = parent::accountNameFilter($acc_path, $limit);
    }
    catch (\Throwable $ex) {
      $this->logger->error(
        "accountNameFilter returned @type @message<br />@fullerror",
        ['@type' => get_class($ex), '@message' => $ex->makeMessage(), '@fullerror'=> $e]
      );
    }
    return $result;
  }

  /**
   * {@inheritDoc}
   * @todo upgrade the API
   */
  public function convertPrice(string $node_path) : \stdClass {

  }

  /**
   * {@inheritDoc}
   */
  public function buildValidateRelayTransaction(BaseTransaction $transaction) : array {
    $additional = parent::buildValidateRelayTransaction($transaction);
    // Convert remote addresses to local wallets.
    foreach($additional as &$entry) {
      $entry->payee = Functions::convertCCaddressToWalletId($entry->payee, $entry->metadata);
      $entry->payer = Functions::convertCCaddressToWalletId($entry->payer, $entry->metadata);

      $entry = (array)$entry;
      $entry['metadata'] = (array)$entry['metadata'];
    }
    return $additional;
  }

  /**
   * Log every request to the network.
   * @param string $endpoint
   * @return \stdClass|NULL
   */
  protected function request(string $endpoint = '/') :\stdClass|NULL {
    $headers = $this->options[RequestOptions::HEADERS] ? array_map( // cumbersome...
      function ($v, $k){return "$k: $v";},
      $this->options[RequestOptions::HEADERS],
      array_keys($this->options[RequestOptions::HEADERS])
    ) : [];
    \Drupal::logger('Credit Commons')->notice(
       Markup::create("@headers<br />@method @domain/@endpoint<br />@body"),
      [
        '@headers' => Markup::create(implode('<br />', $headers)),
        '@method' => strtoupper($this->method),
        '@domain' => $this->baseUrl,
        '@endpoint' => $endpoint,
        '@body' => $this->options[RequestOptions::BODY] ?? ""
      ]
    );
    return parent::request($endpoint);
  }

}
