<?php

namespace Drupal\mcapi_forms;

use Drupal\mcapi\Functions;
use Drupal\system\Entity\Menu;
use Drupal\Core\Entity\EntityForm;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\Element\PathElement;
use Drupal\Core\Link;
use Drupal\Core\Url;

/**
 * Form to edit the FormEditor entity.
 */
class FormEditorForm extends EntityForm {

  /**
   * {@inheritdoc}
   */
  public function form(array $form, FormStateInterface $form_state) {
    $form = parent::form($form, $form_state);
    $w = 0;
    $form['label'] = [
      '#title' => t('Label of the form'),
      '#type' => 'textfield',
      '#default_value' => $this->entity->label(),
      '#size' => 40,
      '#maxlength' => 80,
      '#required' => TRUE,
      '#weight' => $w++
    ];
    $form['id'] = [
      '#type' => 'machine_name',
      '#default_value' => $this->entity->id(),
      '#maxlength' => 12,
      '#machine_name' => [
        'exists' => '\Drupal\mcapi_forms\Entity\McForm::load',
      ],
      '#disabled' => !$this->entity->isNew(),
    ];
    $form['wallet_link_title'] = [
      '#title' => t('Link title to show on wallet'),
      '#description' => t('Link to this form from the wallet. Optionally use token [mc_wallet:name]'),
      '#type' => 'textfield',
      '#default_value' => $this->entity->getWalletLinkTitle(),
      '#weight' => $w++
    ];

    $form['workflow_type'] = [
      '#title' => t('Workflow type'),
      '#description' => t('This also determines the starting state of the transaction.'),
      '#type' => 'select',
      '#options' => Functions::entityOptions('mc_workflow'),
      '#default_value' => $this->entity->getWorkflowId(),
      '#required' => TRUE,
      '#weight' => $w++
    ];
    $form['path'] = [
      '#title' => t('Path'),
      '#description' => t('The url path at which this form will appear. Must begin with / and be unique.'),
      '#type' => 'textfield',
      '#default_value' => $this->entity->getPath(),
      '#placeholder' => '/path/to/form',
      '#pattern' => '\/[a-z\-0-9_/]+',
      '#convert_path' => PathElement::CONVERT_URL,
      '#element_validate' => [[get_class(), 'uniquePath']],
      '#required' => TRUE,
      '#field_prefix' => \Drupal::service('router.request_context')->getCompleteBaseUrl(),
      '#weight' => $w++
    ];
    $form['menu'] = [
      '#title' => t('Menu link'),
      '#type' => 'fieldset',
      '#tree' => TRUE,
      'title' => [
        '#title' => t('Menu link title'),
        '#description' => t('To make tabs, use the format link_title/tab_title.'),
        '#type' => 'textfield',
        '#default_value' => $this->entity->getMenuItem()->title,
        '#weight' => 1,
      ],
      'weight' => [
        '#title' => t('Weight'),
        '#type' => 'weight',
        '#default_value' => $this->entity->getMenuItem()->weight,
        '#description' => t('In the menu, the heavier links will sink and the lighter links will be positioned nearer the top.'),
        '#weight' => 2,
      ],
      'add_transaction_tab' => [
        '#title' => t('Add tab to transactions page'),
        '#descriptions' => t('Tabs are ordered alphabetically'),
        '#type' => 'checkbox',
        '#default_value' => $this->entity->hasTransactionTab(),
        '#weight' => 3,
      ],
      '#weight' => $w++,
    ];

    if (!\Drupal::moduleHandler()->moduleExists('menu_ui')) {
      \Drupal::service('module_installer')->install(['menu_ui']);
    }
    $custom_menus = [];
    foreach (Menu::loadMultiple() as $menu_parent => $menu) {
      $custom_menus[$menu_parent] = $menu->label();
    }
    $form['menu']['parent'] = \Drupal::service('menu.parent_form_selector')
      ->parentSelectElement($settings['menu_parent'] ?? 'main:');
    $form['menu']['parent']['#empty_value'] = '';
    $form['menu']['parent']['#title'] = $this->t("Parent menu item");
    $form['menu']['parent']['#default_value'] = $this->entity->getMenuItem()->parent;
    list($required_fields, $optional_fields) = static::getRequiredFields();
    $required = t(
      'Required fields: @fields.',
      ['@fields' => ' {{ ' . implode(' }}, {{ ', $required_fields) . ' }}']
    );
    $optional = t(
      'Optional fields: @fields.',
      ['@fields' => ' {{ ' . implode(' }}, {{ ', $optional_fields) . ' }}']
    );

    $form['twig'] = [
      '#title' => t('Main form @Twig', ['@Twig' => 'Twig']),
      '#description' => implode(' ', [
        t('Use the following @Twig tokens with HTML & css to design your payment form. Linebreaks will be replaced automatically.', ['@Twig' => 'Twig']),
        $required .' '.$optional . ' ' .
        Link::fromTextAndUrl(
          t('What is @Twig?', ['@Twig' => 'Twig']),
          Url::fromUri('http://twig.sensiolabs.org/doc/templates.html')
        )->toString(),
      ]),
      '#type' => 'textarea',
      '#rows' => 6,
      '#default_value' => str_replace(array('\r\n', '\n', '\r'), "\n", $this->entity->getTwigTemplate()),
      '#element_validate' => [[get_class(), 'validateTwigTemplate']],
      '#required' => TRUE,
      '#weight' => $w++
    ];

    $form['quicklink'] = [
      '#title' => t('Show quicklink on wallet'),
      '#descriptions' => t('Only if the wallet extrafield '),
      '#type' => 'checkbox',
      '#default_value' => $this->entity->getQuicklink(),
      '#weight' => $w++
    ];

    return $form;
  }

  /**
   * {@inheritDoc}
   */
  public function save(array $form, FormStateInterface $form_state) {
    $status = $this->entity->save();

    if ($status == SAVED_UPDATED) {
      $this->messenger()->addStatus(t('Form %label has been updated.', ['%label' => $this->entity->label()]));
    }
    else {
      $this->messenger()->addStatus(t('Form %label has been added.', ['%label' => $this->entity->label()]));
    }
    $form_state->setRedirect('entity.entity_form_display.mc_transaction.default');
  }

  /**
   * Element validation callback ensuring that the given path is unique.
   */
  public static function uniquePath(&$element, $form_state) {
    $path = $form_state->getValue('path');
    $url = Url::fromUserInput($path);
    if ($url->isRouted()) {
      $params = $url->getRouteParameters();
      dpm($url);
      if (isset($params['mc_workflow']) and $params['mc_workflow'] <> $form_state->getFormObject()->entity->id()) {
        $form_state->setError($element, t('The path %path already exists.', ['%path' =>  $path]));
      }
    }
  }


  /**
   * Check the required transaction form elements are either preset OR in Twig.
   */
  public static function validateTwigTemplate(array $element, $form_state) {
    $twig_template = $element['#value'];
    list($required_fields, $optional_fields) = static::getRequiredFields();
    preg_match_all('/\{\{ ?([a-z_]+) ?\}\}/', $twig_template, $matches);
    $used_tokens = $matches[1] ?? [];

    // Check for duplicate tokens.
    if (count($used_tokens) <> count(array_unique($used_tokens))) {
      $form_state->setError($element, t('Each twig token can only appear once in the form.'));
    }

    // Check for unknown tokens.
    foreach ($used_tokens as $fname) {
      if (!in_array($fname, $required_fields)) {
        $form_state->setError($element, t('Invalid Twig token @token', ['@token' => '{{ '.$fname.' }}']));
      }
    }

    // Check for missing required tokens.
    foreach ($required_fields as $fname) {
      if (!in_array($fname, $used_tokens)) {
        if (isset($field_definitions[$fname]) and !$field_definitions[$fname]->getDefault()) {
          $form_state->setError($element, t('Missing required token @token', ['@token' => '{{ '.$fname.' }}']));
        }
        elseif (!isset($field_definitions[$fname])) {
          $form_state->setError($element, t('Missing required token @token', ['@token' => '{{ '.$fname.' }}']));
        }
      }
    }

  }
  /**
   * Get the transaction entity's required and optional form fields
   * @return array
   */
  private static function getRequiredFields() : array {
    $required_fields = ['quant', 'description', 'payer', 'payee'];
    $optional_fields = [];
    // Fields that cannot be set
    $fixed = ['xid', 'uuid', 'serial', 'creator', 'workflow', 'state', 'changed', 'created', 'entries'];
    $field_definitions = \Drupal::service('entity_field.manager')
      ->getFieldDefinitions('mc_transaction', 'mc_transaction');
    foreach ($field_definitions as $id => $field_def) {
      /** @var \Drupal\Core\Field\BaseFieldDefinition $field_def */
      if (in_array($id, $fixed)) {
        continue;
      }
      if ($field_def->isRequired()) {
        $required_fields[] = $id;
      }
      else {
        $optional_fields[] = $id;
      }
    }
    return [$required_fields, $optional_fields];
  }

}
