<?php

namespace Drupal\mcapi_limits;

use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\Element;

/**
 * Form constructor for setting a wallet's limits.
 */
class SettingsForm extends ConfigFormBase {

  /**
   * {@inheritDoc}
   */
  public function getFormId() {
    return 'limit_warning_form';
  }

  public function getEditableConfigNames() {
    return ['mcapi_limits.settings'];
  }

  /**
   * {@inheritDoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $config = $this->config('mcapi_limits.settings');
    // This wasn't designed to be an ajax form
    // The fallback plugin still produces a message, so we're putting it here.
    $plugin_name = $form_state->getValue('plugin') ?? $config->get('plugin') ?? 'balanced';
    $plugin_settings = (array)$form_state->getValue(['plugin_settings', $plugin_name]) ?? $config->get('plugin_settings.'.$plugin_name);

    $plugin = \Drupal::service('plugin.manager.mcapi_limits')
      ->createInstance($plugin_name, $plugin_settings);
    $form['plugin'] = [
      '#title' => t('Plugin'),
      '#description' => t('How will the limits be determined?'),
      '#type' => 'select',
      '#options' => \Drupal::service('plugin.manager.mcapi_limits')->optionsList(),
      '#default_value' => $plugin->getPluginId(),
      '#weight' => 1,
      '#ajax' => [
        'callback' => [get_class(), 'ajaxSubform'],
        'wrapper' => 'limits-plugin-settings',
        'effect' => 'fade',
      ]
    ];
    // Show the options form for the selected plugin.
    $form['plugin_settings'] = [
      $plugin->getPluginId() => $plugin->buildConfigurationForm($form, $form_state),
      '#prefix' => '<div id="limits-plugin-settings">',
      '#suffix' => '</div>',
      '#tree' => TRUE,
      '#weight' => 2,
      '#theme_wrappers' => ['form_element'],
      '#attached' => ['library' =>  ['mcapi/worth.element']]
    ];

    $form['skip'] = [
      '#title' => $this->t('Skip balance limit check for the following transactions'),
      '#description' => $this->t('Especially useful for mass transactions and automated transactions'),
      '#type' => 'checkboxes',
      // Casting it here saves us worrying about the default, which is awkward.
      '#default_value' => array_keys(array_filter((array)$config->get('skip'))),
      // Would be nice if this was pluggable, but not needed for now.
      '#options' => [
        'auto' => $this->t("Automated transactions (such as periodic fees)"), // where user = 0 and type = 3rdparty
        'admin' => $this->t("Created by a user with 'manage mcapi' permission"),
        'all' => $this->t("Skip all transactions (warn only)")
      ],
      '#states' => [
        'invisible' => [
          ':input[name="limits[limits_callback]"]' => ['value' => 'limits_none'],
        ],
      ],
      '#weight' => 6,
    ];
        // To move this to the mcapi_mass module we would have to ensure the correct weight of modules
    if (\Drupal::moduleHandler()->moduleExists('mcapi_mass')) {
      $form['skip']['#options']['mass'] = $this->t("Mass transactions"); // NB mass transactions are now a separate module
    }

    $form['override'] = [
      '#title' => $this->t('Allow individual wallet overrides'),
      '#description' => $this->t("Set on each wallet's edit page."),
      '#type' => 'checkbox',
      '#default_value' => $config->get('override'),
      '#weight' => 10
    ];
    $form['warning'] = [
      '#title' => t('Notification that balance limits were crossed.'),
      '#type' => 'fieldset',
      '#tree' => 1,
      '#weight' => 15,
      'threshhold' => [
        '#title' => $this->t('Warning threshhold'),
        '#description' => $this->t('Fraction of the total limit to be reached before triggering a warning. Set to 0 for no warnings. 99 to warn at latest possible moment.'),
        '#type' => 'number',
        '#default_value' => $config->get('warning.threshhold')??0,
        '#weight' => 0,
        '#min' => 0,
        '#max' => 99,
        '#field_suffix' => '%',
        '#size' => 2
      ],
      'subject' => [
        '#title' => t('Subject'),
        '#type' => 'textfield',
        '#default_value' => $config->get('warning.mail.subject'),
        '#maxlength' => 180,
        '#weight' => 1,
        '#states' => [
          'required' => [
            ':input[name="warning[threshhold]"]' => ['!value' => '0']
          ],
          'visible' => [
            ':input[name="warning[threshhold]"]' => ['!value' => '0']
          ]
        ]
      ],
      'body' => [
        '#title' => t('Body'),
        '#description' => t("The [user] tokens, represent the email recipient."),
        '#type' => 'textarea',
        '#default_value' => $config->get('warning.mail.body'),
        '#rows' => 8,
        '#weight' => 2,
        '#states' => [
          'required' => [
            ':input[name="warning[threshhold]"]' => ['!value' => '0']
          ],
          'visible' => [
            ':input[name="warning[threshhold]"]' => ['!value' => '0']
          ]
        ]
      ]
    ];
    if (\Drupal::moduleHandler()->moduleExists('token')) {
      $form['tokens'] = [
        '#theme' => 'token_tree_link',
        '#token_types' => ['current-user', 'user', 'transgression', ],
        '#global_types' => FALSE,
        '#weight' => 20
      ];
    }
    return parent::buildForm($form, $form_state);
  }

  /**
   * {@inheritDoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    \Drupal::service('plugin.manager.mcapi_limits')
      ->createInstance($form_state->getValue('plugin'))
      ->validateConfigurationForm($form, $form_state);
  }

  /**
   * {@inheritDoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $this->config('mcapi_limits.settings')
      ->set('plugin', $form_state->getValue('plugin'))
      ->set('plugin_settings', $form_state->getValue('plugin_settings'))
      ->set('warning.threshhold', $form_state->getValue(['warning', 'threshhold']))
      ->set('warning.mail.subject', $form_state->getValue(['warning', 'subject']))
      ->set('warning.mail.body', $form_state->getValue(['warning', 'body']))
      ->set('override', $form_state->getValue('override'))
      ->set('skip', $form_state->getValue('skip'))
      ->save();
    // Rendered wallets may be affected by changes in balance limits.
    \Drupal\Core\Cache\Cache::invalidateTags(['mc_wallet_view']);
  }


  /**
   * Ajax callback.
   */
  function ajaxSubform($form, $form_state) {
    return $form['plugin_settings'];
  }

}

