<?php

namespace Drupal\mcapi_mass;

use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Entity\ContentEntityForm;
use Drupal\Core\Routing\RouteMatchInterface;
use Drupal\Core\Mail\MailManagerInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\Logger\LoggerChannel;

/**
 * Form builder for multiple payments between one and many wallets.
 */
class MassPayForm extends ContentEntityForm {

  const MASSINCLUDE = 0;
  const MASSEXCLUDE = 1;

  /**
   * @var Drupal\Core\Mail\MailManagerInterface
   */
  protected $mailManager;

  /**
   * Either '12many' or 'many21'
   * @var string
   */
  protected $direction;

  /**
   * Mail template
   * @var array
   */
  protected $configfactory;

  /**
   * The payer and payee field names.
   * @var string
   */
  protected $manyfieldName;
  protected $onefieldName;

  /**
   * A list of all the wallets involved in the mass transaction.
   * @var int[]
   */
  protected $involved = [];

  /**
   * The step of the form. either 1 or 2.
   * @var int
   */
  protected $step;

  /**
   * the logger object
   * @var Drupal\Core\Logger\LoggerChannel
   */
  protected $logger;

  /**
   *
   * @param type $entity_repository
   * @param type $entity_type_bundle_info
   * @param type $time
   * @param MailManagerInterface $mail_manager
   * @param RouteMatchInterface $route_match
   * @param ConfigFactoryInterface $config_factory
   * @param LoggerChannel $logger
   */
  public function __construct($entity_repository, $entity_type_bundle_info, $time, MailManagerInterface $mail_manager, RouteMatchInterface $route_match, ConfigFactoryInterface $config_factory, LoggerChannel $logger) {
    parent::__construct($entity_repository, $entity_type_bundle_info, $time);
    $this->mailManager = $mail_manager;
    $this->direction = $route_match->getRouteObject()->getOption('direction');
    $this->configFactory = $config_factory;
    $this->logger = $logger;

    if ($this->direction == '12many') {
      $this->onefieldName = PAYER_FIELDNAME;
      $this->manyfieldName = PAYEE_FIELDNAME;
    }
    else {
      $this->onefieldName = PAYEE_FIELDNAME;
      $this->manyfieldName = PAYER_FIELDNAME;
    }
  }

  /**
   * {@inheritDoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity.repository'),
      $container->get('entity_type.bundle.info'),
      $container->get('datetime.time'),
      $container->get('plugin.manager.mail'),
      $container->get('current_route_match'),
      $container->get('config.factory'),
      $container->get('logger.channel.mcapi')
    );
  }

  /**
   * Provide an ID for the transaction form consistent with the keys of hook_transction_form_list().
   */
  function transactionFormId() {
    return 'mass_'.$this->direction;
  }


  /**
   * {@inheritDoc}
   */
  public function form(array $form, FormStateInterface $form_state) {
    $this->init($form_state);

    if (count($this->involved) < 2) {
      $this->step = 1;
      $this->step1($form, $form_state);
    }
    else {
      $this->step = 2;
      $form_state->setValidationComplete();
      foreach ($this->entity->flatten() as $t) {
        $transactions[] = $t;
      }
      $form['preview'][] = $this->entityTypeManager
        ->getViewBuilder('mc_transaction')
        ->viewMultiple($transactions, 'sentence');
    }
    return $form;
  }

  /**
   * {@inheritDoc}
   */
  function actionsElement(array $form, FormStateInterface $form_state) {
    $element = parent::actionsElement($form, $form_state);
    if ($this->step == 1) {
      $element['submit']['#value'] = $this->t('Preview');
    }
    else {
      $element['submit']['#value'] = $this->t('Confirm');
    }
    return $element;
  }

  /**
   * {@inheritDoc}
   */
  public function step1(array &$form, FormStateInterface $form_state) {
    $display = $this->getFormDisplay($form_state);
    $form['#parents'] = [];
#    $display->getRenderer($this->manyfieldName)->cardinalitySetMultiple();
    // Build the default transaction form
    foreach ($display->getComponents() as $name => $options) {
      if (!in_array($name, ['workflow', 'created'])) {
        if ($widget = $display->getRenderer($name)) {
          $items = $this->entity->get($name);
          $items->filterEmptyItems();
          $form[$name] = $widget->form($items, $form, $form_state);
          $form[$name]['#access'] = $items->access('edit');
        }
      }
    }
    $form[$this->manyfieldName]['widget']['#title']  = $this->manyfieldName == PAYER_FIELDNAME ? t('Payers') : t('Payees');
    unset($form[$this->manyfieldName]['widget']['#description']);
    unset($form[$this->onefieldName]['widget'][0]['target_id']['#description']);

    // Should probably set the weights of everything, since the existing ones
    // come from the default form display
    $form['description']['#weight'] = 5;

    $form['invert'] = [
      '#type' => 'radios',
      '#required' => TRUE,
    ];
    $form['invert']['#options'] = [
      SELF::MASSINCLUDE => $this->t('The named wallets'),
      SELF::MASSEXCLUDE => $this->t("All wallets except those named"),
    ];

    // Some cosmetic alterations
    if ($this->direction == '12many') {
      $form[$this->onefieldName]['widget']['target_id']['#title'] = $this->t('The one payer');
      $form['invert']['#title'] = $this->t('Will pay');
    }
    else {
      $form[$this->onefieldName]['widget']['target_id']['#title'] = $this->t('The one payee');
      $form['invert']['#title'] = $this->t('Will receive from');
    }
    $form[$this->onefieldName]['#weight'] = 1;
    $form['invert']['#weight'] = 2;
    $form[$this->manyfieldName]['#weight'] = 3;
    $form['worth']['#weight'] = 4;
    $form['description']['#weight'] = 5;
    $form[$this->manyfieldName]['#tags'] = TRUE;
    unset($form[$this->manyfieldName]['widget']['target_id']['#description']);
    unset($form[$this->onefieldName]['widget']['target_id']['#description']);

    $mail_setting  = $this->configFactory->get('mcapi_mass.settings')->get('last_mail');
    //@todo use rules for this
    $form['notification'] = [
      '#title' => $this->t('Notify all parties', [], array('context' => 'accounting')),
      // @todo decide whether to put rules in a different module
      '#description' => $this->moduleHandler->moduleExists('rules') ?
      $this->t('N.B. Ensure this mail does not clash with mails sent by the rules module.') : '',
      '#type' => 'fieldset',
      '#open' => TRUE,
      '#weight' => 20,
      'subject' => [
        '#title' => $this->t('Subject'),
        '#type' => 'textfield',
        // This needs to be stored per-exchange.
        '#default_value' => $mail_setting['subject'],
      ],
      'body' => [
        '#title' => $this->t('Message'),
        // @todo the tokens?
        '#description' => $this->t('The following pseudo-tokens are available: [user:name]'),
        '#type' => 'textarea',
        '#default_value' => $mail_setting['body'],
        '#weight' => 1,
      ],
    ];
  }

  /**
   * {@inheritDoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    if (($triggering_element = $form_state->getTriggeringElement()) && isset($triggering_element['#ajax']['callback'])) {
      return;
    }
    if ($this->step == 1) {
      if ($form_state->getValue('invert') == Self::MASSINCLUDE and !$form_state->getValue($this->manyfieldName)) {
        $form_state->setError(
          $form[$this->manyfieldName],
          $this->t("'@field' is required", ['@field' => $form[$this->manyfieldName]['#title']])
        );
      }

      // Unlike normal one-step entity forms, save the entiry here for step 2
      $this->entity = parent::validateForm($form, $form_state);
      // And save the mail
      $form_state->set('mail', [
        'subject' => $form_state->getValue('subject'),
        'body' => $form_state->getValue('body')
      ]);
      // Reload the form for the confirmation page.
      $form_state->setRebuild(TRUE);
    }
  }

  /**
   * {@inheritDoc}
   */
  public function buildEntity(array $form, FormStateInterface $form_state) {
    $entity = parent::buildEntity($form, $form_state);
    $entity->workflow->target_id = 'admin';
    $entity->manyField = $this->manyfieldName;
    foreach ($entity->{$this->manyfieldName}->getValue() as $item) {
      $theMany[] = $item['target_id'];
    }
    // Excluding wallets
    if ($form_state->getValue('invert') == SELF::MASSEXCLUDE) {
      $field_definition = $this->entity->get($this->manyfieldName)->getFieldDefinition();
      $entity->many = \Drupal::service('plugin.manager.entity_reference_selection')
        ->getSelectionHandler($field_definition) // WalletSelection handler
        ->inverse($theMany);
    }
    // Including wallets
    else {
      $entity->many = $theMany;
    }
    // Set the main transaction participant to the first of many.
    $entity->{$this->manyfieldName}->setValue(array_pop($entity->many));

    $this->involved = array_unique(array_merge($theMany, [$entity->payer->target_id, $entity->payee->target_id]));
    return $entity;
  }


  /**
   * {@inheritDoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    //NOT calling the parent because we don't want to build the entity again
    $form_state->cleanValues();
    $this->updateChangedTime($this->entity);
    $this->configFactory->getEditable('mcapi.settings')
      ->set('masspay_mail', $form_state->get('mail'))
      ->save();
  }

  /**
   * {@inheritDoc}
   */
  public function save(array $form, FormStateInterface $form_state) {
    parent::save($form, $form_state);
    // @todo send in a batch
    if ($params = $form_state->get('mail')) {
      $this->notifyMany();

    }
    // Go to the canonical transaction page
    $form_state->setRedirect(
      'entity.mc_transaction.canonical',
      ['mc_transaction' => $this->entity->id()]
    );

    $this->logger->notice(
      'User @uid created a mass transactions with @num entries :link',
      [
        '@uid' => $this->currentUser()->id(),
        '@num' => $this->entity->children->count() - 1,
        ':link' => $this->entity->toUrl()
      ]
    );
  }

  function notifyMany() {
    $params['mc_transaction'] = $this->entity;
    $batch['title'] = t('Mailing participants');
    foreach (array_chunk($this->many, 50) as $chunk) {
      $batch['operations'][] = ['mass_pay_mail', [$chunk, $this->entity->id()]];
    }
    batch_set($batch);
  }

  /**
   * Provide an ID for the transaction form consistent with the keys of hook_transction_form_list().
   */
  function transactionFormId() {
    return 'mass_'.$this->direction;
  }


}
