<?php

namespace Drupal\mcapi\Form;

use Drupal\views\Entity\View;
use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Cache\Cache;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Form\FormStateInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Form builder for miscellaneous settings.
 */
class MiscSettings extends ConfigFormBase {

  private $moduleHandler;
  private $entityTypeManager;
  protected $config;

  /**
   * {@inheritDoc}
   */
  public function getFormId() {
    return 'mcapi_misc_settings';
  }

  /**
   * Constructor.
   *
   * @param ConfigFactoryInterface $configFactory
   *   The Configfactory service.
   * @param ModuleHandlerInterface $module_handler
   *   The ModuleHandler service.
   * @param EntityTypeManagerInterface $entity_type_manager
   *   The EntityTypeManager service.
   */
  public function __construct(ConfigFactoryInterface $configFactory, ModuleHandlerInterface $module_handler, EntityTypeManagerInterface $entity_type_manager) {
    parent::__construct($configFactory);
    $this->config = $this->config('mcapi.settings');// editable via ConfigFormBaseTrait.
    $this->moduleHandler = $module_handler;
    $this->entityTypeManager = $entity_type_manager;
  }

  /**
   * Injections.
   */
  static public function create(ContainerInterface $container) {
    return new static(
      $container->get('config.factory'),
      $container->get('module_handler'),
      $container->get('entity_type.manager')
    );
  }

  /**
   * {@inheritDoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $i = 0;
    $help[] = $this->t('E.g. for loaves put 0 loaves; for dollars and cents put $0.99; for Hours, Minutes put Hr0:59/4mins.');
    $help[] = $this->t('The main unit is always a zero');
    $help[] = $this->t('The optional divisor /n at the end will render the divisions as a dropdown field showing intervals, in the example, of 15 mins.');
    $help[] = $this->t('Special effects can be accomplised with HTML & css.');
    $form['currency'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Currency'),
      'name' => [
        '#title' => $this->t('Name'),
        '#description' => $this->t("Plural form works best."),
        '#type' => 'textfield',
        '#default_value' => $this->config->get('currency')['name'],
        '#size' => 20,
        '#maxlength' => 128,
        '#weight' => $i++,
      ],
      'format' => [
        '#title' => $this->t('Display format'),
        '#description' => implode(' ', $help),
        '#type' => 'currency_format',
        '#default_value' => implode($this->config->get('currency')['format']), // array
        '#size' => 20,
        '#weight' => $i++
      ],
      'zero_snippet' => [
        '#title' => $this->t('Zero snippet'),
        '#description' => $this->t("If zero transactions are allowed, display them with this html snippet"),
        '#type' => 'textfield',
        '#default_value' => $this->config->get('currency')['zero_snippet'],
        '#size' => 20,
        '#maxlength' => 128,
        '#weight' => $i++,
      ]
    ];
    $token_description = \Drupal::moduleHandler()->moduleExists('token') ?
      $this->t('Template to display a transaction as a single sentence, using tokens.') :
      $this->t('Enable the token module to see available tokens.');

    $form['sentence_template'] = [
      '#title' => $this->t('Transaction sentence template'),
      '#description' => $token_description,
      '#type' => 'textfield',
      '#default_value' => $this->config->get('sentence_template'),
      '#weight' => $i++
    ];

    if ($this->moduleHandler->moduleExists('views')) {
      $form['user_transaction_view'] = [
        '#title' => $this->t("Transaction display"),
        '#description' => $this->t('When looking at a wallet (or user).'),
        '#type' => 'radios',
        '#options' => [
          '' => $this->t('None'),
          'mc_income_expenditure' => $this->t('Income & expenditure'),
          'mc_wallet_transactions' => $this->t('Statement', [], ['context' => 'like a bank statement']),
        ],
        '#default_value' => $this->config->get('user_transaction_view'),
        '#weight' => $i++,
        '#disabled' => !$this->moduleHandler->moduleExists('views')
      ];
    }

    $form['notify'] = [
      '#title' => $this->t('Notify transactees', [], ['context'=>'The other party to the transaction.']),
      '#description' => $this->t('Email a user when someone makes a transaction with them.'),
      '#type' => 'checkbox',
      '#default_value' => !empty($this->config->get('mail_notification.subject')) and !empty($this->config->get('mail_notification.body')),
      '#weight' => $i++,
    ];

    $form['mail_notification'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Notification mail'),
      '#description' => $this->t('Leave blank to NOT send'),
      '#tree' => TRUE,
      'subject' => [
        '#title' => $this->t('Notification subject'),
        '#description' => $token_description,
        '#type' => 'textfield',
        '#default_value' => $this->config->get('mail_notification.subject'),
        '#weight' => 1,
        '#states' => [
          'required' => [
            [':input[name="notify"]' => ['checked' => TRUE]]
          ]
        ]
      ],
      'body' => [
        '#title' => $this->t('Mail body'),
        '#description' => $token_description,
        '#type' => 'textarea',
        '#weight' => $i++,
        '#default_value' => $this->config->get('mail_notification.body'),
        '#weight' => 2,
        '#states' => [
          'required' => [
            [':input[name="notify"]' => ['checked' => TRUE]]
          ]
        ]
      ],
      '#weight' => $i++,
      '#states' => [
        'visible' => [
          [':input[name="notify"]' => ['checked' => TRUE]]
        ]
      ]
    ];

    $form['notify_signatories'] = [
      '#title' => $this->t('Notify signatories', [], ['context' => 'Users who should sign a transaction.']),
      '#description' => $this->t('Email a user when a transaction requires their signature.'),
      '#type' => 'checkbox',
      '#default_value' => !empty($this->config->get('signature_notification.subject')) and !empty($this->config->get('signature_notification.body')),
      '#weight' => $i++,
    ];

    // The mail template is REQURED if any of the type->notify flags is checked.
    $form['signature_notification'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Signatures mail'),
      '#tree' => TRUE,
      '#description' => $this->t('Leave blank to NOT send'),
      'subject' => [
        '#title' => $this->t('Notification subject'),
        '#type' => 'textfield',
        '#default_value' => $this->config->get('signature_notification.subject'),
        '#weight' => 1,
        '#states' => [
          'required' => [
            [':input[name="notify_signatories"]' => ['checked' => TRUE]]
          ]
        ]
      ],
      'body' => [
        '#title' => $this->t('Mail body'),
        '#description' => $this->t("SHOULD include the token [xaction:sign-link]"),
        '#type' => 'textarea',
        '#weight' => $i++,
        '#default_value' => $this->config->get('signature_notification.body'),
        '#weight' => 2,
        '#states' => [
          'required' => [
            [':input[name="notify_signatories"]' => ['checked' => TRUE]]
          ]
        ]
      ],
      '#states' => [
        'visible' => [
          [':input[name="notify_signatories"]' => ['checked' => TRUE]]
        ]
      ],
      '#weight' => $i++
    ];

    if (\Drupal::moduleHandler()->moduleExists('token')) {
      $form['mail_template']['tokens'] = [
        '#theme' => 'token_tree_link',
        '#token_types' => ['user', 'xaction', 'site', 'current-date'],
        '#global_types' => FALSE,
        '#weight' => $i++
      ];
    }
    $form['rebuild_xaction_index'] = [
      '#title' => $this->t('Rebuild index'),
      '#description' => $this->t('The index table stores the transactions in an views-friendly format. It should never need rebuilding on a working system.'),
      '#type' => 'fieldset',
      '#weight' => $i++,
      'button' => [
        '#type' => 'submit',
        '#value' => $this->t('Rebuild transaction index'),
        '#submit' => [
          [get_class($this), 'rebuildXactionIndex'],
        ],
      ],
    ];
    return parent::buildForm($form, $form_state);
  }

  /**
   * {@inheritDoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $values = $form_state->getValues();
    $this->config
      ->set('currency', [
        'zero_snippet' => $values['zero_snippet'],
        'name' => $values['name'],
        'format' => $values['format']
      ])
      ->set('mail_notification', $values['mail_notification'])
      ->set('sentence_template', $values['sentence_template'])
      ->set('signature_notification', $values['signature_notification']);
;
    if ($this->moduleHandler->moduleExists('views')) {
      $this->config->set('user_transaction_view', $values['user_transaction_view']);
      $this->userTransactionView($values['user_transaction_view']);
    }
    $this->config->save();
    parent::submitForm($form, $form_state);
    Cache::invalidateTags(
      ['mc_wallet_values', 'mc_wallet_view']
    );
  }

  /**
   * Submit callback.
   *
   * Rebuild the Transaction index table.
   */
  public static function rebuildXactionIndex(array &$form, FormStateInterface $form_state) {
    \Drupal::entityTypeManager()->getStorage('mc_transaction')->indexRebuild();
    \Drupal::messenger()->addStatus(t("Index table now tallies with transaction table."));
    $form_state->setRedirect('system.status');
  }

  /**
   * {@inheritDoc}
   */
  protected function getEditableConfigNames() {
    return ['mcapi.settings'];
  }

  private static function userTransactionView(string $view_name) {
    $inex = View::load('mc_income_expenditure');
    $statement = View::load('mc_wallet_transactions');
    $statement->setStatus(FALSE);
    $inex->setStatus(FALSE);
    if ($view_name == 'mc_income_expenditure') {
      $inex->setStatus(TRUE);
    }
    elseif ($view_name == 'mc_wallet_transactions') {
      $inex->setStatus(FALSE);
    }
    $inex->save();
    $statement->save();
  }

}
