<?php

namespace Drupal\mcapi\Form;

use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\Entity\EntityFieldManager;
use Drupal\Core\Entity\EntityTypeManager;
use Drupal\Core\Entity\ContentEntityType;
use Drupal\Core\Entity\EntityTypeBundleInfo;
use Drupal\Component\Utility\Html;
use Drupal\dynamic_entity_reference\Plugin\Field\FieldType\DynamicEntityReferenceItem;
use Drupal\Core\Field\Plugin\Field\FieldType\EntityReferenceItem;
use Drupal\Core\Entity\EntityReferenceSelection\SelectionPluginManager;

/**
 * Form to edit wallet holder settings.
 */
class WalletSettings extends FormBase {

  /**
   * The wallet settings.
   */
  protected $holderDefinition;
  protected $baseFieldOverrideStorage;
  protected $contentEntityTypes;
  protected $bundleInfo;

  /**
   *
   * @param EntityTypeManager $entity_type_manager
   * @param EntityFieldManager $entity_field_manager
   * @param EntityTypeBundleInfo $entity_type_bundle_info
   */
  public function __construct(EntityTypeManager $entity_type_manager, EntityFieldManager $entity_field_manager, EntityTypeBundleInfo $entity_type_bundle_info) {
    $this->holderDefinition = $entity_field_manager->getBaseFieldDefinitions('mc_wallet')['holder'];
    $this->baseFieldOverrideStorage = $entity_type_manager->getStorage('base_field_override');
    $this->bundleInfo = $entity_type_bundle_info;
    $this->contentEntityTypes = static::getWalletable($entity_type_manager->getDefinitions());
  }

  /**
   * {@inheritDoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('entity_field.manager'),
      $container->get('entity_type.bundle.info')
    );
  }

  /**
   * {@inheritDoc}
   */
  public function getFormId() {
    return 'wallet_holder_settings';
  }

  /**
   * {@inheritDoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state): array {
    $form = [];
    foreach ($this->contentEntityTypes as $entity_type_id) {
      $entity_type = \Drupal::entityTypeManager()->getDefinition($entity_type_id);
      $form[$entity_type_id] = $this->targetTypeFieldSettingsForm($form, $form_state, $entity_type_id);
      $form[$entity_type_id]['handler']['#title'] = t('Reference type for @target_type', ['@target_type' => $entity_type->getLabel()]);
    }

    return $form;
  }

  protected function targetTypeFieldSettingsForm(array $form, FormStateInterface $form_state, $target_type) {

    /** @var \Drupal\dynamic_entity_reference\SelectionPluginManager $manager */
    $manager = \Drupal::service('plugin.manager.dynamic_entity_reference_selection');
    // Get all selection plugins for this entity type.
    $selection_plugins = $manager->getSelectionGroups($target_type);
    $handlers_options = [];
    foreach (array_keys($selection_plugins) as $selection_group_id) {
      if (array_key_exists($selection_group_id, $selection_plugins[$selection_group_id])) {
        $handlers_options[$selection_group_id] = Html::escape($selection_plugins[$selection_group_id][$selection_group_id]['label']);
      }
      elseif (array_key_exists($selection_group_id . ':' . $target_type, $selection_plugins[$selection_group_id])) {
        $selection_group_plugin = $selection_group_id . ':' . $target_type;
        $handlers_options[$selection_group_plugin] = Html::escape($selection_plugins[$selection_group_id][$selection_group_plugin]['base_plugin_label']);
      }
    }
    $form = [
      '#type' => 'container',
      '#process' => [[EntityReferenceItem::class, 'fieldSettingsAjaxProcess']],
      '#element_validate' => [[
          WalletSettings::class,
          'fieldSettingsFormValidate',
      ]],
    ];
    $form['handler'] = [
      '#type' => 'details',
      '#title' => t('Reference type'),
      '#open' => TRUE,
      '#tree' => TRUE,
      '#process' => [[EntityReferenceItem::class, 'formProcessMergeParent']],
    ];

    $form['handler']['handler'] = [
      '#type' => 'select',
      '#title' => t('Reference method'),
      '#options' => $handlers_options,
#      '#default_value' => $field_settings[$target_type]['handler'],
      '#required' => TRUE,
      '#ajax' => TRUE,
      '#limit_validation_errors' => [],
    ];
    $form['handler']['handler_submit'] = [
      '#type' => 'submit',
      '#value' => t('Change handler'),
      '#limit_validation_errors' => [],
      '#attributes' => [
        'class' => ['js-hide'],
      ],
      //'#submit' => ['entity_reference_settings_ajax_submit'],
    ];

    $form['handler']['handler_settings'] = [
      '#type' => 'containesr',
      '#attributes' => ['class' => ['entity_reference-settings']]
    ]+$manager->getSelectionHandler($this->holderDefinition, NULL, $target_type)
      ->buildConfigurationForm([], $form_state);

    return $form;
  }

  /**
   * Form element validation handler; Stores the new values in the form state.
   *
   * @param array $element
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   */
  public static function fieldSettingsFormValidate(array $element, FormStateInterface $form_state) {
    $types = self::getWalletable(\Drupal::entityTypeManager()->getDefinitions());
    foreach ($types as $target_type) {
      if ($form_state->getValue(['handler_settings','target_bundles']) === []) {
        $form_state->setValue(['handler_settings', 'target_bundles'], NULL);
      }
      $form_state->unsetValue(['handler_settings', 'target_bundles_update']);
      $form_state->unsetValue(['handler_submit',]);
    }
  }


  /**
   * {@inheritDoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
die('submit');
    $field_storage_definition = $this->holderDefinition->getFieldStorageDefinition('holder');
    //$field_definition = $this->holderDefinition->getFieldDefinition('holder');

    // Copy the field properties to a new configuration entity.
    $this->baseFieldOverrideStorage->create([
      'field_name' => 'title',
      'entity_type_id' => 'node',
      'bundle' => 'article',
      'settings' => $field_definition->getSettings(),
      'third_party_settings' => $field_definition->getThirdPartySettings(),
      'cardinality' => $field_storage_definition->getCardinality(),
      'translatable' => $field_storage_definition->isTranslatable(),
    ])->save();


    $form_state->setRedirect(
      'entity.mc_wallet.canonical',
      ['mc_wallet' => $this->entity->id()]
    );
    parent::save($form, $form_state);
  }

  static function getWalletable(array $etm_definitions) : array {
    $types = [];
    $ignore = [
      'path_alias', 'mc_transaction', 'mc_wallet', 'block_content', 'menu_link_content', 'shortcut', 'taxonomy_term', 'comment'
    ];
    foreach ($etm_definitions as $def) {
      if (!in_array($def->id(), $ignore)) {
        if ($def instanceof ContentEntityType) {
          $types[] = $def->id();
        }
      }
    }
    return $types;
  }

}
